//+------------------------------------------------------------------+
//|                                                OpenCL_Sample.mq5 |
//|                        Copyright 2013, MetaQuotes Software Corp. |
//|                                              http://www.mql5.com |
//+------------------------------------------------------------------+
#property copyright "Copyright 2013, MetaQuotes Software Corp."
#property link      "http://www.mql5.com"
#property version   "1.00"
//---       
#property script_show_inputs
//+------------------------------------------------------------------+
//|                               |
//+------------------------------------------------------------------+
enum rgb_color
  {
   red,
   green,
   blue
  };
//---  
input float     InpXStart=-21.0f; //    X
input float     InpYStart=-21.0f; //    Y
input float     InpXStep=0.1f;    //    X
input float     InpYStep=0.1f;    //    Y
input float     InpXFinish=21.0f; //    X
input float     InpYFinish=21.0f; //    Y
input rgb_color InpColor=green;   //  
//---    OpenCL
const string cl_src=
                    "__kernel void Func(float x_start,                     \r\n"
                    "                   float y_start,                     \r\n"
                    "                   float x_step,                      \r\n"
                    "                   float y_step,                      \r\n"
                    "                   __global float *out)               \r\n"
                    "{                                                     \r\n"
                    "  uint  i = get_global_id(0);                         \r\n"
                    "  uint  j = get_global_id(1);                         \r\n"
                    "  uint  w = get_global_size(1);                       \r\n"
                    "  float x = x_start+i*x_step;                         \r\n"
                    "  float y = y_start+j*y_step;                         \r\n"
                    "                                                      \r\n"
                    "  float a;                                            \r\n"
                    "  if(y!=0) a = x/y;                                   \r\n"
                    "  else     a = (float)0xFFFFFF;                       \r\n"
                    "                                                      \r\n"
                    "  out[i*w+j] = pow(sin(pow(x,2)+pow(y,2)+atan(a)),2); \r\n"
                    "}                                                     \r\n"
                    "                                                      \r\n"
                    "__kernel void Grad(float min,                         \r\n"
                    "                   float dif,                         \r\n"
                    "                   uint c1,                           \r\n"
                    "                   uint c2,                           \r\n"
                    "                   __global float *in,                \r\n"
                    "                   __global uint *out)                \r\n"
                    "{                                                     \r\n"
                    "  uint i = get_global_id(0);                          \r\n"
                    "  uint x = round(255*(in[i]-min)/dif);                \r\n"
                    "  out[i] = c1*(x/16)+c2*(x%16);                       \r\n"
                    "}                                                     \r\n";
//+------------------------------------------------------------------+
//| Script program start function                                    |
//+------------------------------------------------------------------+
void OnStart()
  {
//---      
   uint const_1=0;
   uint const_2=0;
//---        
   PrepareColorConstants(InpColor,const_1,const_2);
//---     
   uint w=(uint)((InpXFinish-InpXStart)/InpXStep);
   uint h=(uint)((InpYFinish-InpYStart)/InpYStep);
   uint size=w*h;
//---     
   float values1[];
   float values2[];
   uint  colors1[];
   uint  colors2[];
//---    
   ArrayResize(values1,size);
   ArrayResize(colors1,size);
   ArrayResize(values2,size);
   ArrayResize(colors2,size);
//---   OpenCL
   Print("\n  OpenCL:");
   WithoutOpenCL(values1,colors1,w,h,size,const_1,const_2);
//---   OpenCL
   Print("\n  OpenCL:");
   WithOpenCL(values2,colors2,w,h,size,const_1,const_2);
//---      
//---   
   string obj_name="OpenCL_"+IntegerToString(ChartID());
//---  
   string res_name="BmpLbl_"+IntegerToString(ChartID());
//---   
   ObjectCreate(0,obj_name,OBJ_BITMAP_LABEL,0,0,0);
   ObjectSetInteger(0,obj_name,OBJPROP_XDISTANCE,0);
   ObjectSetInteger(0,obj_name,OBJPROP_YDISTANCE,0);
   ResourceCreate(res_name,colors1,w,h,0,0,w,COLOR_FORMAT_XRGB_NOALPHA);
   ObjectSetString(0,obj_name,OBJPROP_BMPFILE,"::"+res_name);
//---  
   ChartRedraw();
//---  10 
   Sleep(10000);
//---  
   ObjectDelete(0,obj_name);
  }
//+------------------------------------------------------------------+
//|     OpenCL                   |
//+------------------------------------------------------------------+
void WithoutOpenCL(float &values[],uint &colors[],const uint w,const uint h,
                   const uint size,const uint const_1,const uint const_2)
  {
//---    
   uint x=GetTickCount();
//---   
   for(uint i=0;i<h;i++)
      for(uint j=0;j<w;j++)
         values[i*w+j]=Func(InpXStart+i*InpXStep,InpYStart+j*InpYStep);
//---     
   Print("   = "+IntegerToString(GetTickCount()-x)+" ");
//---   ,    
//---      
   float min,dif;
   GetMinAndDifference(values,size,min,dif);
//---    
   x=GetTickCount();
//---     
   uint a;
   for(uint i=0;i<size;i++)
     {
      a=(uint)MathRound(255*(values[i]-min)/dif);
      colors[i]=const_1*(a/16)+const_2*(a%16);
     }
//---       
   Print("   = "+IntegerToString(GetTickCount()-x)+" ");
  }
//+------------------------------------------------------------------+
//|     OpenCL                    |
//+------------------------------------------------------------------+
void WithOpenCL(float &values[],uint &colors[],const uint w,const uint h,
                const uint size,const uint const_1,const uint const_2)
  {
//---    OpenCL
   int cl_ctx;
   int cl_prg;
   int cl_krn_1;
   int cl_krn_2;
   int cl_mem_1;
   int cl_mem_2;
//---    OpenCL ( )
   if((cl_ctx=CLContextCreate(CL_USE_ANY))==INVALID_HANDLE)
     {
      Print("OpenCL not found");
      return;
     }
//---        cl_src
   if((cl_prg=CLProgramCreate(cl_ctx,cl_src))==INVALID_HANDLE)
     {
      CLContextFree(cl_ctx);
      Print("OpenCL program create failed");
      return;
     }
//---         
   if((cl_krn_1=CLKernelCreate(cl_prg,"Func"))==INVALID_HANDLE)
     {
      CLProgramFree(cl_prg);
      CLContextFree(cl_ctx);
      Print("OpenCL kernel_1 create failed");
      return;
     }
//---       
   if((cl_krn_2=CLKernelCreate(cl_prg,"Grad"))==INVALID_HANDLE)
     {
      CLKernelFree(cl_krn_1);
      CLProgramFree(cl_prg);
      CLContextFree(cl_ctx);
      Print("OpenCL kernel_2 create failed");
      return;
     }
//---  OpenCL    
   if((cl_mem_1=CLBufferCreate(cl_ctx,size*sizeof(float),CL_MEM_READ_WRITE))==INVALID_HANDLE)
     {
      CLKernelFree(cl_krn_2);
      CLKernelFree(cl_krn_1);
      CLProgramFree(cl_prg);
      CLContextFree(cl_ctx);
      Print("OpenCL buffer create failed");
      return;
     }
//---  OpenCL    
   if((cl_mem_2=CLBufferCreate(cl_ctx,size*sizeof(uint),CL_MEM_READ_WRITE))==INVALID_HANDLE)
     {
      CLBufferFree(cl_mem_1);
      CLKernelFree(cl_krn_2);
      CLKernelFree(cl_krn_1);
      CLProgramFree(cl_prg);
      CLContextFree(cl_ctx);
      Print("OpenCL buffer create failed");
      return;
     }
//---    
   uint x=GetTickCount();
//---   ,       
   uint offset[2]={0,0};
//---   ,     
   uint work[2];
   work[0]=h;
   work[1]=w;
//---   
//---    
   CLSetKernelArg(cl_krn_1,0,InpXStart);
   CLSetKernelArg(cl_krn_1,1,InpYStart);
   CLSetKernelArg(cl_krn_1,2,InpXStep);
   CLSetKernelArg(cl_krn_1,3,InpYStep);
   CLSetKernelArgMem(cl_krn_1,4,cl_mem_1);
//---   
   CLExecute(cl_krn_1,2,offset,work);
//---     
   CLBufferRead(cl_mem_1,values);
//---     
   Print("   = "+IntegerToString(GetTickCount()-x)+" ");
//---   ,    
//---      
   float min,dif;
   GetMinAndDifference(values,size,min,dif);
//---    
   x=GetTickCount();
//---    
   uint offset2[1]={0};
   uint work2[1];
   work2[0]=size;
//---     
//---    
   CLSetKernelArg(cl_krn_2,0,min);
   CLSetKernelArg(cl_krn_2,1,dif);
   CLSetKernelArg(cl_krn_2,2,const_1);
   CLSetKernelArg(cl_krn_2,3,const_2);
   CLSetKernelArgMem(cl_krn_2,4,cl_mem_1);
   CLSetKernelArgMem(cl_krn_2,5,cl_mem_2);
//---   
   CLExecute(cl_krn_2,1,offset2,work2);
//---     
   CLBufferRead(cl_mem_2,colors);
//---       
   Print("   = "+IntegerToString(GetTickCount()-x)+" ");
//---   OpenCL
   CLBufferFree(cl_mem_1);
   CLBufferFree(cl_mem_2);
   CLKernelFree(cl_krn_1);
   CLKernelFree(cl_krn_2);
   CLProgramFree(cl_prg);
   CLContextFree(cl_ctx);
  }
//+------------------------------------------------------------------+
//|                                           |
//+------------------------------------------------------------------+
float Func(const float x,const float y)
  {
//---    
   float arg;
   if(y!=0) arg=x/y;
   else     arg=(float)0xFFFFFF;
//---     
   return((float)MathPow(MathSin(MathPow(x,2)+MathPow(y,2)+MathArctan(arg)),2));
  }
//+------------------------------------------------------------------+
//|             |
//+------------------------------------------------------------------+
void PrepareColorConstants(const rgb_color clr,uint &const_1,uint &const_2)
  {
//---   
   switch(InpColor)
     {
      case red:
         const_1=1048576;
         const_2=65536;
         break;
      case green:
         const_1=4096;
         const_2=256;
         break;
      case blue:
         const_1=16;
         const_2=1;
         break;
     }
  }
//+------------------------------------------------------------------+
//|    ,       |
//|                         |
//+------------------------------------------------------------------+
void GetMinAndDifference(const float &arr[],const int size,float &min,float &dif)
  {
//---    
   dif=-FLT_MAX; //     
   min=FLT_MAX;  // 
   for(int i=0;i<size;i++)
     {
      //---  
      if(arr[i]>dif)
         dif=arr[i];
      //---  
      if(arr[i]<min)
         min=arr[i];
     }
//---  
   dif=dif-min;
  }
//+------------------------------------------------------------------+
